@@ -17,6 +17,9 @@
 #include "opengl.h"
 #include "vertex-buffer.h"
 #endif
+#ifdef HAVE_LIBRESVG
+#include "resvg-port.h"
+#endif
 
 #ifdef IMPLEMENT_FREETYPE_GL
 #include "platform.c"
@@ -31,6 +34,9 @@
 #include "vertex-attribute.c"
 #include "vertex-buffer.c"
 #endif
+#ifdef HAVE_LIBRESVG
+#include "resvg-port.c"
+#endif
 #endif
 
 #ifdef __cplusplus
