@@ -1,7 +1,7 @@
 ;;; gforth.el --- major mode for editing (G)Forth sources
 
 ;; Authors: Bernd Paysan, Anton Ertl, David Kühling
-;; Copyright (C) 1995,1996,1997,1998,2000,2001,2003,2004,2007,2008,2010,2011,2012,2013,2014,2015,2016,2017,2018,2019,2020,2021,2022,2023 Free Software Foundation, Inc.
+;; Copyright (C) 1995,1996,1997,1998,2000,2001,2003,2004,2007,2008,2010,2011,2012,2013,2014,2015,2016,2017,2018,2019,2020,2021,2022,2023,2024 Free Software Foundation, Inc.
 
 ;; This file is part of Gforth.
 
@@ -181,16 +181,22 @@ PARSED-TYPE specifies what kind of text is parsed. It should be
 	(("]" "]l") definition-starter (font-lock-keyword-face . 1))
 	((":") definition-starter (font-lock-keyword-face . 1)
 	 "[ \t\n]" t name (font-lock-function-name-face . 3))
+	(("?:") definition-starter (font-lock-keyword-face . 1)
+	 "[ \t\n]" t name (font-lock-function-name-face . 3))
+	((":is") definition-starter (font-lock-keyword-face . 1)
+	 "[ \t\n]" t name (font-lock-function-name-face . 3))
+	((":method") definition-starter (font-lock-keyword-face . 1)
+	 "[ \t\n]" t name (font-lock-function-name-face . 3))
 	(("to:") definition-starter (font-lock-keyword-face . 1)
 	 "[ \t\n]" t name (font-lock-function-name-face . 3))
 	(("defer@:") definition-starter (font-lock-keyword-face . 1)
 	 "[ \t\n]" t name (font-lock-function-name-face . 3))
 	(("event:") definition-starter (font-lock-keyword-face . 1)
 	 "[ \t\n]" t name (font-lock-function-name-face . 3))
-	(("immediate" "compile-only" "restrict")
+	(("immediate" "compile-only" "restrict" "addressable" "addressable:")
 	 immediate (font-lock-keyword-face . 1))
 	(("does>") definition-starter (font-lock-keyword-face . 1))
-	((":noname" "comp:" "compsem:" "opt:" "defer:" "defer@-opt:" "to-opt:") definition-starter (font-lock-keyword-face . 1))
+	((":noname" "comp:" "compsem:" "opt:" "fold1:") definition-starter (font-lock-keyword-face . 1))
 	((":trigger-on(") definition-starter (font-lock-keyword-face . 1)
 	 "[)]" t comment (font-lock-function-name-face . 3))
 	((";" ";code" ";abi-code") definition-ender (font-lock-keyword-face . 1))
@@ -253,7 +259,7 @@ PARSED-TYPE specifies what kind of text is parsed. It should be
 	(("postpone" "[is]" "defers" "[']" "[compile]") 
 	 compile-only (font-lock-keyword-face . 2)
 	 "[ \t\n]" t name (font-lock-function-name-face . 3))
-	(("is" "what's") immediate (font-lock-keyword-face . 2)
+	(("is" "what's" "action-of") immediate (font-lock-keyword-face . 2)
 	 "[ \t\n]" t name (font-lock-function-name-face . 3))
 	(("<is>" "'" "see") non-immediate (font-lock-keyword-face . 2)
 	 "[ \t\n]" t name (font-lock-function-name-face . 3))
@@ -265,7 +271,8 @@ PARSED-TYPE specifies what kind of text is parsed. It should be
 	 "[ \t\n]" t name (font-lock-variable-name-face . 3))
 
 	(("create" "variable" "constant" "2variable" "2constant" "fvariable"
-	  "$variable" "$[]variable" "fconstant" "value" "2value" "fvalue"
+	  "$variable" "$[]variable" "fconstant" "value" "2value"
+	  "fvalue" "varue" "2varue" "fvarue"
 	  "field" "user" "vocabulary" "cs-vocabulary" "create-interpret/compile"
 	  "interpret/compile:" "debug:" "field:" "2field:" "ffield:"
 	  "sffield:" "dffield:" "uvar" "uvalue" "cfield:" "wfield:" "lfield:"
@@ -278,6 +285,8 @@ PARSED-TYPE specifies what kind of text is parsed. It should be
 	(("method" "umethod")
 	 non-immediate (font-lock-type-face . 2)
 	 "[ \t\n]" t name (font-lock-function-name-face . 3))
+	(("with") compile-only (font-lock-type-face . 2)
+	 "[ \t\n]" t name (font-lock-variable-name-face . 3))
 	("\\S-+%" non-immediate (font-lock-type-face . 2))
 	(("defer" "alias" "create-interpret/compile:") 
 	 non-immediate (font-lock-type-face . 1)
@@ -299,6 +308,10 @@ PARSED-TYPE specifies what kind of text is parsed. It should be
 	 "[ \t\n]" t name (font-lock-function-name-face . 3)
 	 "[ \t\n]" t name (font-lock-function-name-face . 3)
 	 "[\n]" nil comment (font-lock-variable-name-face . 3))
+	(("c-weak-function" "c-value") non-immediate (font-lock-type-face . 1)
+	 "[ \t\n]" t name (font-lock-function-name-face . 3)
+	 "[ \t\n]" t name (font-lock-function-name-face . 3)
+	 "[\n]" nil comment (font-lock-variable-name-face . 3))
 	(("c-callback") non-immediate (font-lock-type-face . 1)
 	 "[ \t\n]" t name (font-lock-function-name-face . 3)
 	 "[\n]" nil comment (font-lock-variable-name-face . 3))
@@ -314,9 +327,9 @@ PARSED-TYPE specifies what kind of text is parsed. It should be
 	 immediate (font-lock-constant-face . 3))
 	("\\([&#]-?[0-9a-f.]+\\|\\(0x-?\\|\\$-?\\)[0-9a-f.]+\\|%-?[01]+\\)"
 	 immediate (font-lock-constant-face . 3))
-	("\"[^\"]**" immediate (font-lock-string-face . 1)
-	 "[\"\n]" nil string (font-lock-string-face . 1))
-	("\".*\""
+	("\"[[:cntrl:][:nonascii:][:multibyte:]!#-~ ]*?" immediate (font-lock-string-face . 1)
+	 "[^\\]\"" nil string (font-lock-string-face . 1))
+	("\"[[:cntrl:][:nonascii:][:multibyte:]!#-~ ]*?\""
 	 immediate (font-lock-string-face . 3))
 	("[a-z\-0-9]+(" immediate (font-lock-comment-face . 1)
 	 "[)]" nil comment (font-lock-comment-face . 1))
@@ -371,8 +384,7 @@ PARSED-TYPE specifies what kind of text is parsed. It should be
     (("ptr" "asptr" "[]") 
      immediate (font-lock-keyword-face . 2)
      "[ \t\n]" t name (font-lock-variable-name-face . 3))
-    (("class;" "how:" "self" "new" "new[]" "definitions" "class?" "with"
-      "endwith")
+    (("class;" "how:" "self" "new" "new[]" "definitions" "class?" "endwith")
      non-immediate (font-lock-keyword-face . 2))
     (("object") non-immediate (font-lock-type-face . 2)))
   "Hilighting description for words of the \"OOF\" package")
@@ -428,9 +440,9 @@ INDENT1 and INDENT2 are indentation specifications of the form
 	  "[if]" "[ifdef]" "[ifundef]" "[begin]" "[for]" "[do]" "[?do]" "[:"
 	  "[n:l" "[n:h" "[n:d" "[d:l" "[d:h" "[d:d" "[f:l" "[f:h" "[f:d" "[{:")
 	 (0 . 2) (0 . 2))
-	((":" ":noname" "code" "abi-code" "struct" "m:" ":m" "class" "uclass" 
+	((":" ":is" ":method" "?:" ":noname" "code" "abi-code" "struct" "m:" ":m" "class" "uclass" "with" 
 	  "interface" "c-library" "c-library-name" "comp:" "opt:" "post:"
-	  "begin-structure" "extend-structure" "event:" "to-opt:" "defer@-opt:" "to:" "defer@:" ":trigger-on(")
+	  "begin-structure" "extend-structure" "event:" "fold1:" "to:" "defer@:" ":trigger-on(")
 	 (0 . 2) (0 . 2) non-immediate)
 	("\\S-+%$" (0 . 2) (0 . 0) non-immediate)
 	((";" ";m") (-2 . 0) (0 . -2))
@@ -438,7 +450,7 @@ INDENT1 and INDENT2 are indentation specifications of the form
 	  "[then]" "[endif]" "[loop]" "[+loop]" "[next]" 
 	  "[until]" "[again]" "loop" ";]" "nope")
 	 (-2 . 0) (0 . -2))
-	(("end-code" "end-class" "end-interface" "end-class-noname" 
+	(("end-code" "end-class" "endwith" "end-interface" "end-class-noname" 
 	  "end-interface-noname" "end-struct" "class;" "end-c-library" "end-structure")
 	 (-2 . 0) (0 . -2) non-immediate)
 	(("protected" "public" "how:") (-1 . 1) (0 . 0) non-immediate)
@@ -447,7 +459,10 @@ INDENT1 and INDENT2 are indentation specifications of the form
 	 (-2 . 2) (0 . 0))
 	(("does>" "compile>" "int>" ";code" ";abi-code") (-1 . 1) (0 . 0))
 	(("while" "[while]") (-2 . 4) (0 . 2))
-	(("repeat" "[repeat]") (-4 . 0) (0 . -4))))
+	(("repeat" "[repeat]") (-4 . 0) (0 . -4))
+	(("{{" "vt{{") (0 . 2) (0 . 2) immediate)
+	(("}}h" "}}v" "}}z" "}}vp" "}}p" "}}vt") (-2 . 0) (-2 . 0) immediate)
+	))
 
 (defvar forth-local-indent-words nil 
   "List of Forth words to prepend to `forth-indent-words', when a forth-mode
@@ -746,7 +761,7 @@ End:\" construct).")
 ;;;
 (defvar forth-defining-words 
   '("VARIABLE" "CONSTANT" "2VARIABLE" "2CONSTANT" "FVARIABLE" "FCONSTANT"
-   "USER" "VALUE" "2Value" "field" "end-struct" "VOCABULARY" "CREATE" ":" "CODE"
+   "USER" "VALUE" "2Value" "field" "end-struct" "VOCABULARY" "CREATE" ":" "?:" ":is" ":method" "CODE"
    "DEFER" "ALIAS" "interpret/compile:" "debug:" "field:" "2field:" "ffield:"
    "sffield:" "dffield:" "uvar" "uvalue" "voctable" "method" "umethod")
   "List of words, that define the following word.
