@@ -18,6 +18,7 @@ install_debian() {
   sudo apt-get -y install libharfbuzz-dev
   sudo apt-get -y install libvulkan-dev
   sudo apt-get -y install libpng-dev
+  sudo apt-get -y install libwebp-dev
   sudo apt-get -y install libfreetype6-dev
   sudo apt-get -y install libgstreamer1.0-dev
   sudo apt-get -y install libgstreamer-plugins-base1.0-dev
@@ -47,7 +48,7 @@ install_alpine() {
     sudo apk add freetype-dev build-base autoconf automake m4 libtool git \
         coreutils gcc libffi-dev mesa-dev glew-dev libx11-dev \
         libxrandr-dev glfw-dev harfbuzz-dev gstreamer-dev gst-plugins-base-dev \
-	opus-dev pulseaudio-dev pipewire-dev wayland-dev unzip texinfo wayland-protocols libxkbcommon-dev
+	opus-dev pulseaudio-dev pipewire-dev wayland-dev unzip texinfo wayland-protocols libxkbcommon-dev libwebp-dev
     (cd /tmp && git clone https://github.com/nothings/stb.git && \
     sudo mkdir /usr/include/stb && sudo cp stb/*.h /usr/include/stb && rm -rf stb)
 }
@@ -59,7 +60,7 @@ install_fedora() {
 	libtool libtool-ltdl libtool-ltdl-devel git \
         coreutils gcc libffi-devel mesa-devel glew-devel libx11-devel \
         libXrandr-devel glfw-devel harfbuzz-devel gstreamer-devel gst-plugins-base-devel wayland-protocols-devel libxkbcommon-devel \
-	opus-devel pulseaudio-devel unzip texinfo
+	opus-devel libwebp-devel pulseaudio-devel unzip texinfo
     (cd /tmp && git clone https://github.com/nothings/stb.git && \
     sudo mkdir /usr/include/stb && sudo cp stb/*.h /usr/include/stb && rm -rf stb)
 }
@@ -71,7 +72,7 @@ install_opensuse() {
     freetype2-devel harfbuzz-devel libpulse-devel libopus-devel \
     libva-devel libva-gl-devel linux-glibc-devel libxkbcommon-devel \
     makeinfo texinfo info wayland-devel wayland-protocols-devel m4 \
-    emacs-nox libffi-devel libX11-devel
+    emacs-nox libffi-devel libX11-devel libwebp-devel
 }
 
 install_linux() {
@@ -91,7 +92,7 @@ install_osx() {
   brew tap forthy42/homebrew-zsh
   brew update > /dev/null
   brew upgrade > /dev/null
-  brew install gcc harfbuzz texinfo xz mesa premake automake yodl
+  brew install gcc harfbuzz texinfo xz mesa premake automake yodl libtool
   export PATH="/usr/local/opt/texinfo/bin:$PATH"
   brew install --cask xquartz mactex
   export PATH="/Library/TeX/texbin:$PATH"
